/*
******************************************************************************

   BLT implementation for X target - fast blitting functions.

   Copyright (C) 2001	Christoph Egger	[Christoph_Egger@t-online.de]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <math.h>

#include "xblt.h"

#include <ggi/display/xcommon.h>
#include <ggi/display/x.h>

#include <X11/X.h>

#include <ggi/ggi.h>



int fast_crossblit(ggiBlt_t src, ggiBlt_t dst)
{
	int rc = GALLOC_OK;
	struct GA_x_bob *bob_src = NULL;
	struct GA_x_bob *bob_dst = NULL;


	if (src == NULL) return GGI_EARGINVAL;
	if (dst == NULL) return GGI_EARGINVAL;

	bob_src = GETBOB(src);
	bob_dst = GETBOB(dst);


#warning TODO: Convert src->w,src->h and dst->w,dst->h into pixels, if they are in dots

	if (src->alpha != NULL) {
		if (dst->alpha == NULL) {
			dst->alpha = calloc(1, dst->w * dst->h * sizeof(unsigned char));
			if (!dst->alpha) {
				rc = GGI_ENOMEM;
				goto exit;
			}	/* if */
		}	/* if */

		memcpy(dst->alpha, src->alpha,
			src->w * src->h * sizeof(unsigned char));
	}	/* if */

	if (src->vis == dst->vis) {
		/* same visual => same pixelformat */

		if ((src->w == dst->w) && (src->h == dst->h)) {
			/* same size => no clipping, no overflow */

			memcpy(bob_dst->bobimage, bob_src->bobimage,
				src->w * src->h * sizeof(ggi_pixel));
		}	/* if */
		rc = GALLOC_OK;
		goto exit;
	}	/* if */

	{
	ggi_pixel *image_src = bob_src->bobimage;
	ggi_pixel *image_dst = bob_dst->bobimage;

	int x, y;
	int ystart, i;


	y = 0;
	ystart = y * src->w;

	for (i = ystart; i <= ((src->w - 1) * (src->h - 1)); ) {
		ggi_color col;

		x = i - ystart;

		if (x >= dst->w) goto cont;

		ggiUnmapPixel(src->vis, image_src[0], &col);
		image_dst[0] = ggiMapColor(dst->vis, &col);

		image_dst++;
cont:
		image_src++;

		i++;
		x++;
		if (x == src->w) {
			/* newline */
			ystart = i;
			y++;
			if (y >= dst->h) break;
		}	/* if */

	}	/* for */

	}



	rc = GALLOC_OK;

exit:
	return rc;
}	/* fast_crossblit */


int fast_crossvis2vis(ggi_visual_t src, int sx, int sy, int sw, int sh,
		      ggi_visual_t dst, int dx, int dy,
		      enum ggiGA_coordbase cm)
{
	if (src == NULL) return GGI_EARGINVAL;
	if (dst == NULL) return GGI_EARGINVAL;

#warning Convert sx,sy, sw,sh and dx,dy into pixels, if they are given in DOTS

	return ggiCrossBlit(src, sx,sy, sw,sh, dst, dx,dy);
}	/* fast_crossvis2vis */


int fast_crossblit2vis(ggiBlt_t blt, ggi_visual_t vis,
		       int dx, int dy, enum ggiGA_coordbase cm)
{
	struct GA_x_bob *bob = NULL;
	const ggi_pixel *bob_buf = NULL;
	ggi_pixel *vis_buf = NULL;
	const unsigned char *alpha = NULL;
	ggi_mode mode;

	int i, y, ystart;

	if (blt == NULL) return GGI_EARGINVAL;

	if (vis == blt->vis) {
		/* same pixel format */
		return fast_blit2vis(blt, dx, dy);
	}	/* if */


	bob = GETBOB(blt);
	bob_buf = bob->bobimage;
	alpha = blt->alpha;

#warning Convert dx, dy, blt->w and blt->h into pixels, if they are given in DOTS

	y = 0;
	ystart = y * blt->w;

	vis_buf = malloc(blt->w * sizeof(ggi_pixel));
	if (!vis_buf) return GGI_ENOMEM;

	ggiGetMode(blt->vis, &mode);

	for (i = ystart; i <= ((blt->w - 1) * (blt->h - 1)); ) {
		ggi_color col;
		int x;

		x = i - ystart;

		ggiUnmapPixel(blt->vis, bob_buf[0], &col);
		if (alpha != NULL) {

			if ((dx + x) >= mode.virt.x) goto cont;
			if ((dy + y) >= mode.virt.y) goto cont;

			switch(alpha[0]) {
			case 0:		/* Opaque */
				vis_buf[x] = ggiMapColor(blt->vis, &col);
				break;

			case 0xff:	/* Transparent */
				ggiGetPixel(blt->vis, dx + x, dy + y,
					&vis_buf[x]);
				break;

			default:	/* Partially Transparent*/
			{
				ggi_color sprcol,bgcol;
				ggi_pixel bgpix;

				ggiGetPixel(blt->vis, dx + x, dy + y,
					    &bgpix);
				ggiUnmapPixel(blt->vis, bob_buf[0], &sprcol);
				ggiUnmapPixel(blt->vis, bgpix, &bgcol);

				sprcol.r = (sprcol.r * (255 - alpha[0])
					+ bgcol.r * alpha[0]) / 255;
				sprcol.g = (sprcol.g * (255 - alpha[0])
					+ bgcol.g * alpha[0]) / 255;
				sprcol.b = (sprcol.b * (255 - alpha[0])
					+ bgcol.b * alpha[0]) / 255;

				vis_buf[x] = ggiMapColor(blt->vis, &sprcol);
			}
			}	/* switch */
cont:

			alpha++;
		} else {
			vis_buf[i - ystart] = ggiMapColor(vis, &col);
		}	/* if */


		i++;
		x++;
		bob_buf++;

		if (x == blt->w) {
			/* newline */
			ggiPutHLine(vis, dx, dy + y, blt->w, vis_buf);

			ystart = i;
			y++;
		}	/* if */

	}	/* for */

	free(vis_buf);

	return GALLOC_OK;
}	/* fast_crossblit2vis */


int fast_blit2vis(ggiBlt_t blt, int dx, int dy)
{
	struct GA_x_bob *bob = NULL;
	enum ggiGA_coordbase cm;


	if (blt == NULL) return GGI_EARGINVAL;

	bob = GETBOB(blt);

	cm = blt->cm;

#warning Convert dx, dy, blt->w and blt->h into pixels, if they are given in DOTS

	if (blt->alpha != NULL) {
		int x,y, i;
		int ystart;
		const ggi_pixel *bob_buf;
		ggi_pixel *vis_buf;
		const unsigned char *alpha;
		ggi_mode mode;


		bob_buf = bob->bobimage;
		alpha = blt->alpha;
		ggiGetMode(blt->vis, &mode);

		y = ystart = 0;

		vis_buf = malloc(blt->w * sizeof(ggi_pixel));
		if (!vis_buf) return GGI_ENOMEM;

		for (i = 0; i <= ((blt->w - 1) * (blt->h - 1));) {
			x = i - ystart;

			if ((dx + x) >= mode.virt.x) goto cont;
			if ((dy + y) >= mode.virt.y) goto cont;

			switch(alpha[0]) {
			case 0:		/* Opaque */
				vis_buf[x] = bob_buf[0];
				break;

			case 0xff:	/* Transparent */
				ggiGetPixel(blt->vis, dx + x, dy + y,
					&vis_buf[x]);
				break;

			default:	/* Partially Transparent */
				{
				ggi_color sprcol,bgcol;
				ggi_pixel bgpix;

				ggiGetPixel(blt->vis, dx + x, dy + y,
					    &bgpix);
				ggiUnmapPixel(blt->vis, bob_buf[0], &sprcol);
				ggiUnmapPixel(blt->vis, bgpix, &bgcol);

				sprcol.r = (sprcol.r * (255 - alpha[0])
					+ bgcol.r * alpha[0]) / 255;
				sprcol.g = (sprcol.g * (255 - alpha[0])
					+ bgcol.g * alpha[0]) / 255;
				sprcol.b = (sprcol.b * (255 - alpha[0])
					+ bgcol.b * alpha[0]) / 255;

				vis_buf[x] = ggiMapColor(blt->vis, &sprcol);
				break;
				}
			}	/* switch */
cont:

			alpha++;
			bob_buf++;

			i++;
			x++;
			if (x == blt->w) {
				/* newline */
				ggiPutHLine(blt->vis, dx, dy + y,
					    blt->w, vis_buf);

				ystart = i;
				y++;
			}	/* if */

		}	/* for */

		free(vis_buf);

	} else {
		ggiPutBox(blt->vis, dx,dy, blt->w,blt->h,bob->bobimage);
	}	/* if */

	return GALLOC_OK;
}	/* fast_blit2vis */
