/*
******************************************************************************

   LibBuf extension stubs.

   Copyright (c) Wed Aug 29 2001 by: 
	Brian S. Julin		bri@tull.umassp.edu

  
   
This software may be redistributed under the terms of the GNU Library
(a.k.a. "Lesser") Public License -- see the file COPYING in the parent 
directory of the source code.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


******************************************************************************
*/

#include <ggi/internal/buf.h>
#include <ggi/internal/buf_debug.h>

ggiBuf_t ggiBufTie(ggi_visual_t vis, ggiGA_resource_handle handle_tank,
                   int narg, void *args)
{

	/* Perform only *very basic* checks here, as we 
	 * wish to leave control of things up to the target sublib
	 * as much as possible.
	 */
	if (vis == NULL) return NULL;
	if (handle_tank == NULL) {
		printf("%s:%s:%i: Cannot tie NULL handle\n", DEBUG_INFO);
		return NULL;
	}	/* if */

	if (ggiGAIsFailed(handle_tank)) {
		printf("%s:%s:%i: Cannot tie failed handle.\n", DEBUG_INFO);
		/* libGAlloc failed! */
		return NULL;
	}	/* if */

	return LIBGGI_BUFEXT(vis)->_tie(vis, NULL, NULL, handle_tank, 
					narg, args);

}	/* ggiBufTie */


int ggiBufUntie(ggiBuf_t *buf)
{
	int rc = GALLOC_OK;
	ggiGA_resource_handle *buf_ga, carb, tank;
	ggi_visual *vis;

	if (buf == NULL) return GGI_EARGINVAL;
	if (*buf == NULL) return GGI_EARGINVAL;

	buf_ga = GA_LL(*buf);
	if (buf_ga == NULL) return GGI_EARGINVAL;

	carb = NULL;
	tank = NULL;

	while (*buf_ga) {
	  if (ggiGAIsCarb(*buf_ga)) carb = *buf_ga;
	  if ((ggiGAGetType(*buf_ga) & GA_RT_TYPE_MASK) == GA_RT_BUFFER) 
	    if (ggiGAIsTank(*buf_ga)) tank = *buf_ga;
	  buf_ga++;
	}

	vis = ggiGAGetVisual(tank);
	if (vis == NULL) return GGI_EARGINVAL;

	rc = LIBGGI_BUFEXT(vis)->_untie(vis, buf, carb, tank);

	buf = NULL;

	LIB_ASSERT(rc == GALLOC_OK);
	return rc;
}	/* ggiBufUntie */


/*-* Stub function code. */
int ggiBufStrip(ggi_visual_t vis)
{
	return LIBGGI_BUFEXT(vis)->strip(vis);
}	/* ggiBufStrip */


int ggiBuf_clad(ggi_visual_t vis, int flags, 
		int numbufs, ggiBuf_t *bufs, ggiBuf_t *res)
{
	return LIBGGI_BUFEXT(vis)->_clad(vis, flags, numbufs, bufs, res);
}	/* ggiBuf_clad */



/*-*/

/* per-buf overloads */

int ggiBufBuf2Mem(ggiBuf_t src, size_t xoffset,
		  size_t yoffset, size_t stride,
		  void *membuf, size_t deststride,
		  size_t linewidth, size_t numlines)
{
	ggi_ll_stops_t stops;
	ggi_ll_ststate_t ststate;

	if (src == NULL) return GGI_EARGINVAL;

	stops = GGI_LL_STOPS(src);
	if (stops == NULL) return GGI_EARGINVAL;
	if (stops->storage2mem == NULL) return GGI_EARGINVAL;

	ststate = GGI_LL_STSTATE(src);
	if (ststate == NULL) return GGI_EARGINVAL;

	if (membuf == NULL) return GGI_EARGINVAL;

	/* TODO: Decide if we want bounds checking here, or in stops. */

	return stops->storage2mem(ststate, xoffset, yoffset,
				  stride, membuf, deststride,
				  linewidth, numlines);
}       /* ggiBufBuf2Mem */


int ggiBufMem2Buf(ggiBuf_t dest, size_t xoffset,
		  size_t yoffset, size_t stride,
		  void *membuf, size_t srcstride,
		  size_t linewidth, size_t numlines)
{
	ggi_ll_stops_t stops;
	ggi_ll_ststate_t ststate;

	if (dest == NULL) return GGI_EARGINVAL;

	stops = GGI_LL_STOPS(dest);
	if (stops == NULL) return GGI_EARGINVAL;
	if (stops->mem2storage == NULL) return GGI_EARGINVAL;

	ststate = GGI_LL_STSTATE(dest);
	if (ststate == NULL) return GGI_EARGINVAL;

	if (membuf == NULL) return GGI_EARGINVAL;

	/* TODO: Decide if we want bounds checking here, or in stops. */

	return stops->mem2storage(ststate, xoffset, yoffset,
				  stride, membuf, srcstride,
				  linewidth, numlines);
}       /* ggiBufMem2Buf */


int ggiBufBuf2Buf(ggiBuf_t dest, size_t xoffset,
		  size_t yoffset, size_t stride,
		  ggiBuf_t src, size_t srcxoffset,
		  size_t srcyoffset, size_t srcstride,
		  size_t linewidth, size_t numlines)
{
	ggi_ll_stops_t stops, dstops;
	ggi_ll_ststate_t ststate, dststate;

	if (src == NULL) return GGI_EARGINVAL;

	stops = GGI_LL_STOPS(src);
	if (stops == NULL) return GGI_EARGINVAL;

	ststate = GGI_LL_STSTATE(src);
	if (ststate == NULL) return GGI_EARGINVAL;

	if (dest == NULL) return GGI_EARGINVAL;

	dstops = GGI_LL_STOPS(dest);
	if (dstops == NULL) return GGI_EARGINVAL;
	if (dstops->storage2storage == NULL) return GGI_EARGINVAL;

	dststate = GGI_LL_STSTATE(dest);
	if (dststate == NULL) return GGI_EARGINVAL;

	/* TODO: Decide if we want bounds checking here, or in stops. */

	return dstops->storage2storage(dststate, xoffset, yoffset, stride,
				       ststate, stops,
				       srcxoffset, srcyoffset, srcstride,
				       linewidth, numlines);
}       /* ggiBufBuf2Buf */


int ggiBufConst2Buf(ggiBuf_t dest, size_t xoffset,
		    size_t yoffset, size_t stride,
		    ggi_bufval val, size_t linewidth, size_t numlines)
{
	ggi_ll_stops_t stops;
	ggi_ll_ststate_t ststate;

	if (dest == NULL) return GGI_EARGINVAL;

	stops = GGI_LL_STOPS(dest);
	if (stops == NULL) return GGI_EARGINVAL;
	if (stops->const2storage == NULL) return GGI_EARGINVAL;

	ststate = GGI_LL_STSTATE(dest);
	if (ststate == NULL) return GGI_EARGINVAL;

	/* Do we want bounds checking here? */
	return stops->const2storage(ststate, xoffset, 
				    yoffset, stride, val,
				    linewidth, numlines);
}       /* ggiBufConst2Buf */

