/*
******************************************************************************

   Demo of the OVL extension.

   Authors:	2001 Christoph Egger	[Christoph_Egger@t-online.de]
  
   This code is placed in the public domain and may be used freely for any
   purpose.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>

#include <ggi/ggi.h>
#include <ggi/ovl.h>


static int do_demo(ggi_visual_t vis)
{
	int rc;
	ggiOvl_t ovl_sprite_ptr;
	ggiOvl_t ovl_sprite_cursor;

	ggi_color col;
	ggi_mode *mode;
	int width, height;
	int inc, temp;
	char *pr;

	struct ggiGA_resource_props props_motor;
	struct ggiGA_resource_props props_tank;
	struct ggiGA_resource_props props_carb;

	ggiGA_resource_list reqlist = NULL;
	ggiGA_resource_list reslist = NULL;
	ggiGA_resource_handle frame_handle = NULL;
	ggiGA_resource_handle sprite_ptr_motor = NULL;
	ggiGA_resource_handle sprite_ptr_tank = NULL;
	ggiGA_resource_handle sprite_ptr_carb = NULL;
	ggiGA_resource_handle sprite_cursor_motor = NULL;
	ggiGA_resource_handle sprite_cursor_tank = NULL;
	ggiGA_resource_handle sprite_cursor_carb = NULL;


	ggiOvlClearMotorProperties(&props_motor);
	ggiOvlClearTankProperties(&props_tank);
	ggiOvlClearCarbProperties(&props_carb);


	printf("call ggiGAAddSimpleMode...");
	rc = ggiGAAddSimpleMode(vis, &reqlist, GGI_AUTO, GGI_AUTO,
			GGI_AUTO, GT_AUTO, &frame_handle);
	printf("done -- rc: %i\n", rc);



	props_motor.sub.motor.grid_start.x = 0;
	props_motor.sub.motor.grid_start.y = 0;
	props_tank.size.area.x = 50;
	props_tank.size.area.y = 50;
	props_tank.sub.tank.gt = GT_AUTO;

	printf("Add Sprite-Pointer...");
	rc = ggiOvlAdd(&reqlist,
		       &props_motor, &props_carb, &props_tank, NULL, 
		       GA_RT_SPRITE_POINTER,
		       &sprite_ptr_motor, &sprite_ptr_carb, 
		       &sprite_ptr_tank, NULL);
			
	printf("done -- rc: %i, sprite_ptr (motor/tank/carb) %p,%p,%p\n", rc,
		sprite_ptr_motor, sprite_ptr_tank, sprite_ptr_carb);


	props_tank.size.area.x = GGI_AUTO;
	props_tank.size.area.y = GGI_AUTO;
	props_tank.sub.tank.gt = GT_AUTO;

	printf("Add Sprite-Cursor...");
	rc = ggiOvlAdd(&reqlist, &props_motor, &props_carb, &props_tank, NULL,
		       GA_RT_SPRITE_CURSOR,
		       &sprite_cursor_motor, &sprite_cursor_carb, 
		       &sprite_cursor_tank, NULL);
		       
	printf("done -- rc: %i, sprite_cursor (motor/carb/tank) %p,%p,%p\n", 
	       rc,sprite_cursor_motor,sprite_cursor_carb,sprite_cursor_tank);

	printf("call ggiGAanprintf() before ggiGACheck()\n");
	ggiGAanprintf(vis, reqlist, 1024, NULL, &pr);
	printf("%s\n", pr);
	free(pr);

	rc = ggiGACheck(vis, reqlist, &reqlist);
	if (rc != GALLOC_OK) {
		printf("check failed: %i!\n", rc);
		return -1;
	}	/* if */

	printf("call ggiGAanprintf() after ggiGACheck()\n");
	ggiGAanprintf(vis, reqlist, 1024, NULL, &pr);
	printf("%s\n", pr);
	free(pr);

	rc = ggiGASet(vis, reqlist, &reslist);
	if (rc != GALLOC_OK) {
		printf("set failed: %i!\n", rc);
		return -1;
	}	/* if */

	printf("reqlist.....\n");
	ggiGAanprintf(vis, reqlist, 1024, NULL, &pr);
	printf("%s\n", pr);
	free(pr);

	printf("reslist.....\n");
	ggiGAanprintf(vis, reslist, 1024, NULL, &pr);
	printf("%s\n", pr);
	free(pr);

	frame_handle = ggiGAHandle(reslist, reqlist, frame_handle);
	sprite_ptr_motor = ggiGAHandle(reslist, reqlist, sprite_ptr_motor);
	sprite_ptr_carb = ggiGAHandle(reslist, reqlist, sprite_ptr_carb);
	sprite_ptr_tank = ggiGAHandle(reslist, reqlist, sprite_ptr_tank);

	sprite_cursor_motor = 
		ggiGAHandle(reslist, reqlist, sprite_cursor_motor);
	sprite_cursor_carb = ggiGAHandle(reslist, reqlist, sprite_cursor_carb);
	sprite_cursor_tank = ggiGAHandle(reslist, reqlist, sprite_cursor_tank);


	printf("call ggiOvlTie()...\n");
	ovl_sprite_ptr = ggiOvlTie(vis,
				   sprite_ptr_motor,
				   sprite_ptr_carb,
				   sprite_ptr_tank);
	if (!ovl_sprite_ptr) {
		printf("tie failed!\n");
		return -1;
	}	/* if */

	printf("call ggiOvlTie()...\n");
	ovl_sprite_cursor = ggiOvlTie(vis,
				      sprite_cursor_motor,
				      sprite_cursor_carb,
				      sprite_cursor_tank);
	if (!ovl_sprite_ptr) {
		printf("tie failed!\n");
		return -1;
	}	/* if */



	/* Set foreground color to white */
	col.r = 0xffff;
	col.g = 0xffff;
	col.b = 0xffff;
	ggiSetGCForeground(vis, ggiMapColor(vis, &col));

	/* Check what mode we actually got */
	mode = ggiGAGetGGIMode(frame_handle);

	width = mode->visible.x;
	height = mode->visible.y;


	printf("draw bunch of lines...\n");

	/* Draw a bunch of lines... */
	inc = height / 66;
	for (temp = 0; temp <= (height-inc); temp += inc) {
		ggiDrawLine(vis, 0, temp, width-1, height-temp);
	}	/* for */

	/* ...and flush the screen */
	ggiFlush(vis);


	printf("call ggiOvlSetImage()...\n");
	ggiOvlSetImage(ovl_sprite_ptr, vis, 50,50, NULL);

	printf("call ggiOvlShow()...\n");
	ggiOvlShow(ovl_sprite_ptr);
	ggiOvlShow(ovl_sprite_cursor);
	for (;;) {

		static int hide = 0;
		ggi_event ev;

		struct timeval timeout={0,1000};

		ggiEventPoll(vis, emAll, &timeout);
		usleep(timeout.tv_usec);

		while (ggiEventsQueued(vis, emAll)) {
			ggiEventRead(vis, &ev, emAll);

			switch(ev.any.type) {
			case evKeyPress:
			case evKeyRepeat:

				switch(ev.key.sym) {
				case GIIK_Enter:
					hide ^= 1;

					if (hide) {
						ggiOvlHide(ovl_sprite_ptr);
					} else {
						ggiOvlShow(ovl_sprite_ptr);
					}	/* if */
					break;

				case GIIUC_Escape:
					goto cleanup;

				}	/* switch */
			}	/* switch */	
		}	/* while */


		ggiFlush(vis);
	}	/* for */

	ggiOvlHide(ovl_sprite_ptr);
	ggiOvlHide(ovl_sprite_cursor);


cleanup:


	rc = ggiOvlUntie(&ovl_sprite_ptr);
	printf("ret-code from ggiOvlUntie: %i\n", rc);

	rc = ggiGAReleaseList(vis, &reqlist);
	printf("ret-val from ggiGAReleaseList: %i\n", rc);
	printf("reqlist.....\n");
	ggiGAanprintf(vis, reqlist, 1024, NULL, &pr);
	printf("%s", pr);
	free(pr);

	printf("Final current resource list.....\n");
	ggiGAEmptyList(&reqlist);
	ggiGAGet(vis, &reqlist);
	ggiGAanprintf(vis, reqlist, 1024, NULL, &pr);
	printf("%s", pr);
	free(pr);

	ggiGAEmptyList(&reqlist);

	return 0;
}	/* do_demo */



int main(void)
{
	ggi_visual_t vis;

	/* Initialize LibGGI */
	if (ggiInit() != 0) {
		fprintf(stderr, "Unable to initialize LibGGI\n");
		exit(1);
	}	/* if */
	
	/* Initialize OVL extension */
	if (ggiOvlInit() != 0) {
		ggiPanic("Unable to initialize LibOVL extension\n");
	}	/* if */

	/* Open the default visual */
	if ((vis = ggiOpen(NULL)) == NULL) {
		ggiPanic("Unable to open default visual\n");
	}	/* if */
	
	/* Turn on asynchronous mode (which should always be used) */
	ggiSetFlags(vis, GGIFLAG_ASYNC);
	

	/* Attach the OVL extension to the visual */
	if (ggiOvlAttach(vis, NULL) < 0) {
		ggiPanic("Unable to attach OVL extension to visual\n");
	}	/* if */


	/* Try the API functions */
	do_demo(vis);

	
	/* Detach extension from visual */
	ggiOvlDetach(vis);

	/* Close visual */
	ggiClose(vis);

	/* Deinitialize OVL extension */
	ggiOvlExit();

	/* Deinitialize LibGGI */
	ggiExit();

	return 0;
}	/* main */
